from .digidocservice.service import DigiDocError


class BaseAction(object):
    @classmethod
    def do_action(cls, view, action_kwargs):
        raise NotImplementedError


class NoAction(object):
    @classmethod
    def do_action(cls, view, action_kwargs):
        return {'success': True}


class IdCardPrepareAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        if not view.get_files():
            return {
                'success': False,
                'code': 'MIN_1_FILE',
            }

        # Create signed document
        service.create_signed_document()

        # add all files
        for file in view.get_files():
            service.add_datafile(file.file_name, file.mimetype, service.HASHCODE, file.size, file.content)

        # Call sign
        try:
            response = service.prepare_signature(**action_kwargs)
            response['success'] = True
            return response

        except DigiDocError as e:
            return {
                'success': False,
                'error_code': e.error_code,
                'message': service.ERROR_CODES.get(int(e.error_code), service.ERROR_CODES[100])
            }


class IdCardFinishAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        try:
            return {
                'success': service.finalize_signature(**action_kwargs),
            }

        except DigiDocError as e:
            return {
                'success': False,
                'error_code': e.error_code,
                'message': service.ERROR_CODES.get(int(e.error_code), service.ERROR_CODES[100])
            }


class SignCompleteAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        return service.get_file_data(view.get_files())


class MobileIdSignAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        if not view.get_files():
            return {
                'success': False,
                'code': 'MIN_1_FILE',
            }

        # Create signed document
        service.create_signed_document()

        # add all files
        for file in view.get_files():
            service.add_datafile(file.file_name, file.mimetype, service.HASHCODE, file.size, file.content)

        try:
            # Call sign
            resp = service.mobile_sign(**action_kwargs)

        except DigiDocError as e:
            return {
                'success': False,
                'error_code': e.error_code,
                'message': service.ERROR_CODES.get(int(e.error_code), service.ERROR_CODES[100])
            }

        return {
            'success': True,
            'challenge': resp['ChallengeID'],
        }


class MobileIdStatusAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        status_info = service.get_status_info()

        # If error occured
        if status_info['StatusCode'] not in ['OUTSTANDING_TRANSACTION', 'SIGNATURE', 'REQUEST_OK']:
            return {
                'success': False,
                'code': status_info['StatusCode'],
                'message': service.MID_STATUS_ERROR_CODES[status_info['StatusCode']],
            }

        elif status_info['StatusCode'] == 'OUTSTANDING_TRANSACTION':
            return {
                'success': False,
                'pending': True,
            }

        else:
            return {
                'success': True,
            }


class MobileIdAuthenticateAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        try:
            # Call mobile_authenticate
            resp = service.mobile_authenticate(**action_kwargs)

        except DigiDocError as e:
            return {
                'success': False,
                'error_code': e.error_code,
                'message': service.ERROR_CODES.get(int(e.error_code), service.ERROR_CODES[100])
            }

        view.request.session['mid_person_code'] = resp['UserIDCode']
        view.request.session['mid_common_name'] = resp['UserCN']

        return {
            'success': True,
            'challenge': resp['ChallengeID'],
        }


class MobileIdAuthenticateStatusAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        status_info = service.get_mobile_authenticate_status()

        # If error occured
        if status_info['Status'] not in ['OUTSTANDING_TRANSACTION', 'USER_AUTHENTICATED', 'REQUEST_OK']:
            return {
                'success': False,
                'code': status_info['StatusCode'],
                'message': service.MID_STATUS_ERROR_CODES[status_info['StatusCode']],
            }

        elif status_info['Status'] == 'OUTSTANDING_TRANSACTION':
            return {
                'success': False,
                'pending': True,
            }

        else:
            return {
                'success': True,
            }



# -*- coding: utf-8 -*-
import base64
import hashlib
import io
import re
import sys
from zipfile import ZipFile

from lxml import etree

from django.utils.encoding import force_bytes, force_text


class BDOCException(Exception):
    pass


class HashCodesFileEntry(object):
    def __init__(self, filename, content, size):
        self.filename = filename
        self.content = content
        self.size = size

    def as_xml(self):
        root = etree.Element('file-entry')

        root.attrib['full-path'] = self.filename
        root.attrib['hash'] = self.content
        root.attrib['size'] = str(self.size)

        return root


class HashCodesXml(object):
    def __init__(self, algorithm):
        self.algorithm = algorithm
        self.file_entries = []

    def from_data_files(self, data_files):
        self.file_entries = []

        for data_file in data_files:
            self.file_entries.append(self.convert_data_file_to_file_entry(data_file))

        return self.write()

    def convert_data_file_to_file_entry(self, data_file):
        return HashCodesFileEntry(data_file.filename, self.file_hash(data_file.content), data_file.size)

    def file_hash(self, content):
        return base64.b64encode(getattr(hashlib, self.algorithm)(content).digest())

    def write(self):
        root = etree.Element('hashcodes')

        for file_entry in self.file_entries:
            root.append(file_entry.as_xml())

        return "%s%s" % ('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>', force_text(etree.tostring(root)))


class ZipDataFile(object):
    def __init__(self, filename, content, size):
        self.filename = filename
        self.content = force_bytes(content, errors='surrogateescape')
        self.size = size

    def __str__(self):
        return 'ZipDataFile(filename=%s, size=%s, content[:8]=%s)' % (self.filename, self.size, self.content[:8])


class BdocContainer(object):
    DEFAULT_HASH_ALGORITHM = 'sha256'
    HASH_CODES_FILES_REGEX = r'^META-INF/hashcodes-\w+.xml$'

    HASH_ALGORITHMS = ['sha256', 'sha512']

    @classmethod
    def hash_code(cls, content):
        return base64.b64encode(hashlib.sha256(content).digest())

    def __init__(self, file_data, data_files=None):
        if isinstance(file_data, io.BytesIO):
            raw_data = file_data.getvalue()

        else:
            raw_data = file_data

        if data_files is None:
            data_files = []

        self.file_data = raw_data
        self.data_files = data_files

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def get_temporary_file(self):
        buf = io.BytesIO(self.file_data)

        try:
            return buf, ZipFile(buf, 'a')

        except:
            raise BDOCException('ZipArchive could not be opened')

    def hash_codes_format(self):
        buf, archive = self.get_temporary_file()

        # Remove data files
        buf, archive, data_files = self.delete_data_files(archive)

        # Write hashcode files
        self.write_hash_code_files(archive, data_files)

        # Set comment
        archive.comment = force_bytes(self.__get_container_comment())

        # Commit changes
        archive.close()

        return buf.getvalue()

    def data_files_format(self):
        buf, archive = self.get_temporary_file()

        # Remove hashcode files
        buf, archive, data_files = self.delete_hashcode_files(archive)

        assert self.data_files

        # Write datafiles
        for datafile in self.data_files:
            archive.writestr(datafile.file_name, datafile.content)

        # Commit changes
        archive.close()

        return buf.getvalue()

    def delete_data_files(self, archive):
        return self.__delete_files(archive, lambda filename: self.__is_data_file(filename))

    def delete_hashcode_files(self, archive):
        return self.__delete_files(archive, lambda filename: self.__is_hash_code_file(filename))

    def write_hash_code_files(self, archive, data_files):
        for algorithm in self.HASH_ALGORITHMS:
            filename = "META-INF/hashcodes-%s.xml" % algorithm
            file_data = HashCodesXml(algorithm).from_data_files(data_files)

            archive.writestr(filename, file_data.encode())

    @staticmethod
    def __delete_files(archive, test_func):
        deleted_files = []

        buf = io.BytesIO()
        z_out = ZipFile(buf, 'w')

        for zip_info in archive.infolist():
            content = archive.read(zip_info)

            if test_func(zip_info.filename):
                deleted_files.append(ZipDataFile(zip_info.filename, content, zip_info.file_size))

            else:
                z_out.writestr(zip_info, content)

        return buf, z_out, deleted_files

    @classmethod
    def __is_hash_code_file(cls, file_name):
        return re.match(cls.HASH_CODES_FILES_REGEX, file_name)

    @classmethod
    def __is_data_file(cls, file_name):
        return file_name != 'mimetype' and file_name[:9] != 'META-INF/'

    @classmethod
    def __get_container_comment(cls):
        return 'digidoc %s - python %s, %s' % ('0.0.1', sys.version_info, sys.platform)



import base64
import re

from django.http import HttpResponse, HttpResponseRedirect
from django.utils.encoding import force_text, force_bytes
from django.views.generic import TemplateView, View

from .digidocservice.service import DataFile
from .generic import (MobileIdSignViewMixin, MobileIdStatusViewMixin, DigidocCompleteViewMixin, IdCardPrepareViewMixin,
                      IdCardFinishViewMixin)


class SKTestView(TemplateView):
    template_name = 'esteid/test.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        try:
            files = self.request.session['__ddoc_files']

        except KeyError:
            files = {}

        context.update({
            'files': files,
        })

        return context

    def post(self, request, *args, **kwargs):
        """ This method is only for testing and can
            be used to add some files to the session.

            actions:

                - add_file: Add another file
                - remove_file: Add another file
        """
        try:
            files = request.session['__ddoc_files']

        except KeyError:
            files = {}

        action = request.POST.get('action', 'add_file')

        if action == 'remove_file':
            file_name = request.POST.get('file_name', '')

            n_files = {}
            for key, file in files.items():
                if key != file_name:
                    n_files[key] = file

            files = n_files

        else:
            for key, file in request.FILES.items():
                file_name = file.name

                idx = 1
                while file_name in files.keys():
                    file_name = '%d_%s' % (idx, file_name)
                    idx += 1

                files[file_name] = dict(
                    content=force_text(base64.b64encode(file.read())),
                    content_type=file.content_type,
                    size=file.size,
                )

        request.session['__ddoc_files'] = files

        return HttpResponseRedirect('.')


class TestFilesMixin(object):
    def get_files(self):
        files = []

        for file_name, file in self.request.session['__ddoc_files'].items():
            files.append(
                DataFile(
                    file_name=file_name,
                    mimetype=file['content_type'],
                    content_type=None,
                    size=file['size'],
                    content=base64.b64decode(force_bytes(file['content'])),
                )
            )

        return files


class TestMobileIdSignView(TestFilesMixin, MobileIdSignViewMixin, View):
    def build_action_kwargs(self):
        return dict(
            id_code=self.request.POST.get('id_code', ''),
            phone_nr=self.request.POST.get('phone_nr', ''),
            language=self.request.POST.get('language', None),
        )


class TestMobileIdStatusView(TestFilesMixin, MobileIdStatusViewMixin, View):
    pass


class TestDownloadContainerView(TestFilesMixin, DigidocCompleteViewMixin, View):
    def get(self, request, *args, **kwargs):
        # Get the file data
        the_file = self.do_action(*args, **kwargs)

        # Destroy digidoc session
        self.destroy_digidoc_session()

        # Clear active files
        try:
            del request.session['__ddoc_files']

        except KeyError:
            pass

        # Download the file
        response = HttpResponse(the_file, content_type='application/vnd.bdoc-1.0')
        response['Content-Disposition'] = 'attachment; filename=' + 'signed.bdoc'
        return response


class TestIdCardSignView(TestFilesMixin, IdCardPrepareViewMixin, View):
    def build_action_kwargs(self):
        return {
            'certificate': self.request.POST.get('certificate', ''),
            'token_id': self.request.POST.get('token_id', ''),
        }


class TestIdCardFinishView(TestFilesMixin, IdCardFinishViewMixin, View):
    def build_action_kwargs(self):
        return {
            'signature_id': self.request.POST.get('signature_id', ''),
            'signature_value': self.request.POST.get('signature_value', ''),
        }


class AuthenticationView(TemplateView):
    template_name = 'esteid/authenticate.html'

    def __init__(self, *args, **kwargs):
        self.id_auth = None
        self.id_err = None

    def dispatch(self, request, *args, **kwargs):
        self.id_auth = getattr(request, 'id_auth', None)
        self.id_err = getattr(request, 'id_err', None)

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        self.request.session['id_auth'] = self.id_auth
        context['id_auth'] = self.id_auth
        context['id_err'] = self.id_err

        return context
